@@ -9,7 +9,6 @@ import pandas as pd
 class AverageMeter(object):
     """
     Computes and stores the average and current value
-
     Attributes
     ----------
     val : float
@@ -36,7 +35,6 @@ class AverageMeter(object):
     def update(self, val, n=1):
         """
         Updates current internal state
-
         Parameters
         ----------
         val : float
@@ -50,10 +48,9 @@ class AverageMeter(object):
         self.avg = self.sum / self.count
 
 
-class TorchFitter:
+class TorchFitterBase:
     """
     Helper class to implement a training loop in PyTorch
-
     Parameters
     ----------
     model : torch.nn.Module
@@ -87,7 +84,7 @@ class TorchFitter:
                  step_scheduler=False,
                  folder='models',
                  verbose=True,
-                 log_file=True,
+                 save_log=True,
                  ):
 
         if type(loss) == type:
@@ -98,10 +95,9 @@ class TorchFitter:
         self.epoch = 0  # current epoch
         self.verbose = verbose
 
-        self.base_dir = f'./{folder}'
-        if not os.path.exists(self.base_dir):
-            os.makedirs(self.base_dir)
+        self.base_dir = f'{folder}'
 
+        self.save_log = save_log
         self.log_path = f'{self.base_dir}/log.txt'
         self.best_metric = 0
 
@@ -117,20 +113,33 @@ class TorchFitter:
         self.step_scheduler = step_scheduler  # do scheduler.step after optimizer.step
         self.log(f'Fitter prepared. Device is {self.device}')
 
-    def fit(self, 
-            train_loader, 
-            val_loader=None, 
-            n_epochs=1, 
-            metric=None, 
-            metric_kwargs={}, 
-            early_stopping=0, 
+    def _unpack(self, data):
+        raise NotImplementedError('This class is a base class')
+
+    def reduce_loss(self, loss, weights):
+        # Apply sample weights if existing
+        if len(loss.shape) > 0:
+            # apply weights
+            if weights is not None:
+                loss = loss * torch.unsqueeze(weights, 1)
+
+            # reduction
+            loss = loss.mean()
+        return loss
+
+    def fit(self,
+            train_loader,
+            val_loader=None,
+            n_epochs=1,
+            metric=None,
+            metric_kwargs={},
+            early_stopping=0,
             early_stopping_mode='min',
             early_stopping_alpha=0.0,
             early_stopping_pct=0.0,
             save_checkpoint=True,
             verbose_steps=0):
         """ Fits a model
-
         Parameters
         ----------
         train_loader : torch.utils.data.DataLoader
@@ -155,12 +164,11 @@ class TorchFitter:
             Whether to save the checkpoint when training
         verbose_steps : int, defaults to 0
             number of step to print every training summary
-
         Returns
         -------
         returns a pandas DataFrame object with training history
         """
-        self.best_metric = 10 ** 7 if early_stopping_mode == 'min' else -10 ** 7
+        self.best_metric = np.inf if early_stopping_mode == 'min' else -np.inf
 
         # Use the same train loader for validation. A possible use case is for autoencoders
         if isinstance(val_loader, str) and val_loader == 'training':
@@ -173,7 +181,8 @@ class TorchFitter:
 
             # Update log
             lr = self.optimizer.param_groups[0]['lr']
-            self.log(f'\n{datetime.utcnow().isoformat(" ", timespec="seconds")}\nEPOCH {str(self.epoch+1)}/{str(n_epochs)} - LR: {lr}')
+            self.log(f'\n{datetime.utcnow().isoformat(" ", timespec="seconds")}\n \
+                        EPOCH {str(self.epoch+1)}/{str(n_epochs)} - LR: {lr}')
 
             # Run one training epoch
             t = time.time()
@@ -187,16 +196,20 @@ class TorchFitter:
 
             if val_loader is not None:
                 # Run epoch validation
-                val_summary_loss, calculated_metric = self.validation(val_loader, metric=metric, metric_kwargs=metric_kwargs, verbose_steps=verbose_steps)
+                val_summary_loss, calculated_metric = self.validation(val_loader,
+                                                                      metric=metric,
+                                                                      metric_kwargs=metric_kwargs,
+                                                                      verbose_steps=verbose_steps)
                 history['val'] = val_summary_loss.avg  # validation loss
 
                 # Write log
                 metric_log = f'- metric {calculated_metric},' if calculated_metric else ''
-                self.log(f'\r[RESULT] {(time.time() - t):.2f}s - train loss: {train_summary_loss.avg:.5f} - val loss: {val_summary_loss.avg:.5f} ' + metric_log)
+                self.log(f'\r[RESULT] {(time.time() - t):.2f}s - train loss: {train_summary_loss.avg:.5f} - \
+                           val loss: {val_summary_loss.avg:.5f} ' + metric_log)
 
                 if calculated_metric:
                     history['val_metric'] = calculated_metric
-    
+
                 calculated_metric = calculated_metric if calculated_metric else val_summary_loss.avg
             else:
                 # If no validation is provided, training loss is used as metric
@@ -206,23 +219,24 @@ class TorchFitter:
 
             # Check if result is improved, then save model
             if (((metric) and
-                    (((early_stopping_mode == 'max') and (calculated_metric - max(early_stopping_alpha, es_pct) > self.best_metric)) or
-                     ((early_stopping_mode == 'min') and (calculated_metric + max(early_stopping_alpha, es_pct) < self.best_metric))))
+                    (((early_stopping_mode == 'max') and
+                      (calculated_metric - max(early_stopping_alpha, es_pct) > self.best_metric)) or
+                     ((early_stopping_mode == 'min') and
+                      (calculated_metric + max(early_stopping_alpha, es_pct) < self.best_metric))))
                 or
-                ((metric is None) and (calculated_metric + max(early_stopping_alpha, es_pct) < self.best_metric))):
-
-                    self.log(f'Validation metric improved from {self.best_metric} to {calculated_metric}')
-                    self.best_metric = calculated_metric
-                    self.model.eval()
-                    if save_checkpoint:
-                        savepath = f'{self.base_dir}/best-checkpoint.bin'
-                        self.save(savepath)
-                    es_epochs = 0  # reset early stopping count
+                    ((metric is None) and (calculated_metric + max(early_stopping_alpha, es_pct) < self.best_metric))):
+                self.log(f'Validation metric improved from {self.best_metric} to {calculated_metric}')
+                self.best_metric = calculated_metric
+                self.model.eval()
+                if save_checkpoint:
+                    savepath = f'{self.base_dir}/best-checkpoint.bin'
+                    self.save(savepath)
+                es_epochs = 0  # reset early stopping count
             else:
                 es_epochs += 1  # increase epoch count with no improvement, for early stopping check
 
             # Check if Early Stopping condition is met
-            if (early_stopping > 0) & (es_epochs > early_stopping):
+            if (early_stopping > 0) & (es_epochs >= early_stopping):
                 self.log(f'Early Stopping: {early_stopping} epochs with no improvement')
                 training_history.append(history)
                 break
@@ -239,14 +253,12 @@ class TorchFitter:
     def train_one_epoch(self, train_loader, verbose_steps=0):
         """
         Run one epoch on the train dataset
-
         Parameters
         ----------
         train_loader : torch.DataLoader
             DataLoaders containing the training dataset
         verbose_steps : int, defaults to 0
             number of step to print every training summary
-
         Returns
         -------
         AverageMeter
@@ -266,11 +278,11 @@ class TorchFitter:
                         f'time: {(time.time() - t):.2f} secs, ' +
                         f'ETA: {(len(train_loader)-step)*(time.time() - t)/(step+1):.2f}', end=''
                     )
-            # Calculate batch size
-            batch_size = data['x'].shape[0]
+            # Unpack batch of data
+            x, y, w, batch_size = self._unpack(data)
 
             # Run one batch
-            loss = self.train_one_batch(data)
+            loss = self.train_one_batch(x, y, w)
 
             summary_loss.update(loss.detach().item(), batch_size)
 
@@ -284,40 +296,40 @@ class TorchFitter:
 
         return summary_loss
 
-    def train_one_batch(self, data):
+    def train_one_batch(self, x, y, w=None):
         """
-        Trains one batch of data. 
-        The actions to be done here are: 
+        Trains one batch of data.
+        The actions to be done here are:
         - extract x and y (labels)
         - calculate output and loss
         - backpropagate
-
         Parameters
         ----------
         data : Tuple of torch.Tensor
-            Batch of data. Normally the first element is the data (X) and the second is the label (y), but it dependes on what the Dataset outputs.
-
+            Batch of data. Normally the first element is the data (X) and the second is the label (y),
+            but it depends on what the Dataset outputs.
         Returns
         -------
         torch.Tensor
             A tensor with the calculated loss
         int
-            The batch size 
+            The batch size
         """
-        # extract x and y from the dataloader
-        x = data['x']
-        y = data['y']
-
-        # send them to device
-        x = x.to(self.device).float()
-        y = y.to(self.device)
-
         self.optimizer.zero_grad()
 
         # Output and loss
-        output = self.model(x)
+        if isinstance(x, tuple) or isinstance(x, list):
+            output = self.model(*x)
+        elif isinstance(x, dict):
+            output = self.model(**x)
+        else:
+            output = self.model(x)
+
         loss = self.loss_function(output, y)
 
+        # Reduce loss and apply sample weights if existing
+        loss = self.reduce_loss(loss, w)
+
         # backpropagation
         loss.backward()
 
@@ -326,7 +338,6 @@ class TorchFitter:
     def validation(self, val_loader, metric=None, metric_kwargs={}, verbose_steps=0):
         """
         Validates a model
-
         Parameters
         ----------
         val_loader : torch.utils.data.DataLoader
@@ -337,7 +348,6 @@ class TorchFitter:
             Arguments for the passed metric. Ignored if metric is None
         verbose_steps : int, defaults to 0
             number of step to print every training summary
-
         Returns
         -------
         AverageMeter
@@ -345,7 +355,6 @@ class TorchFitter:
         float
             Calculated metric if a metric is provided, else None
         """
-
         self.model.eval()
         summary_loss = AverageMeter()
         y_preds = []
@@ -358,26 +367,33 @@ class TorchFitter:
                     print(
                         f'\rVal Step {step}/{len(val_loader)}, ' +
                         f'summary_loss: {summary_loss.avg:.5f}, ' +
-                        f'time: {(time.time() - t):.2f} secs,' + 
+                        f'time: {(time.time() - t):.2f} secs,' +
                         f'ETA: {(len(val_loader)-step)*(time.time() - t)/(step+1):.2f}', end=''
                     )
             with torch.no_grad():  # no gradient update
-                x = data['x'].to(self.device).float()
-                y = data['y'].to(self.device)
-                batch_size = x.shape[0]
+                x, y, w, batch_size = self._unpack(data)
 
                 if metric:
                     y_true += y.cpu().numpy().tolist()
 
-                # just forward propagation
-                output = self.model(x)
+                # just forward pass
+                if isinstance(x, tuple) or isinstance(x, list):
+                    output = self.model(*x)
+                elif isinstance(x, dict):
+                    output = self.model(**x)
+                else:
+                    output = self.model(x)
+
                 loss = self.loss_function(output, y)
+
+                # Reduce loss and apply sample weights if existing
+                loss = self.reduce_loss(loss, w)
                 summary_loss.update(loss.detach().item(), batch_size)
 
                 if metric:
                     y_preds += output.cpu().numpy().tolist()
 
-        calculated_metric = metric(y_true, y_preds, **metric_kwargs) if metric else None
+        calculated_metric = metric(y_true, np.argmax(y_preds, axis=1), **metric_kwargs) if metric else None
         metric_log = f'- metric {calculated_metric},' if calculated_metric else ''
         self.log(f'\r[VALIDATION] {(time.time() - t):.2f}s - val. loss: {summary_loss.avg:.5f} ' + metric_log)
         return summary_loss, calculated_metric
@@ -385,14 +401,12 @@ class TorchFitter:
     def predict(self, test_loader, verbose_steps=0):
         """
         Makes predictions using the trained model
-
         Parameters
         ----------
         test_loader : torch.utils.data.DataLoader
             Test Data
         verbose_steps : int, defaults to 0
-            number of step to print every training summary    
-
+            number of step to print every training summary
         Returns
         -------
         np.array
@@ -407,16 +421,20 @@ class TorchFitter:
             if self.verbose & (verbose_steps > 0) > 0:
                 if step % verbose_steps == 0:
                     print(
-                        f'\Prediction Step {step}/{len(test_loader)}, ' +
-                        f'time: {(time.time() - t):.2f} secs,' + 
+                        f'Prediction Step {step}/{len(test_loader)}, ' +
+                        f'time: {(time.time() - t):.2f} secs,' +
                         f'ETA: {(len(test_loader)-step)*(time.time() - t)/(step+1):.2f}', end=''
                     )
             with torch.no_grad():  # no gradient update
-                x = data['x'].to(self.device).float()
-                batch_size = x.shape[0]
+                x, _, _, _ = self._unpack(data)
 
-                # Output and loss
-                output = self.model(x)
+                # Output
+                if isinstance(x, tuple) or isinstance(x, list):
+                    output = self.model(*x)
+                elif isinstance(x, dict):
+                    output = self.model(**x)
+                else:
+                    output = self.model(x)
 
                 y_preds += output.cpu().numpy().tolist()
 
@@ -425,7 +443,6 @@ class TorchFitter:
     def save(self, path, verbose=True):
         """
         Save model and other metadata
-
         Parameters
         ----------
         path : str
@@ -447,12 +464,14 @@ class TorchFitter:
         if self.scheduler is not None:
             data['scheduler_state_dict'] = self.scheduler.state_dict()
 
+        if not os.path.exists(self.base_dir):
+            os.makedirs(self.base_dir)
+
         torch.save(data, path)
 
     def load(self, path):
         """
         Load model and other metadata
-
         Parameters
         ----------
         path : str
@@ -461,28 +480,27 @@ class TorchFitter:
         checkpoint = torch.load(path)
         self.model.load_state_dict(checkpoint['model_state_dict'])
         self.optimizer.load_state_dict(checkpoint['optimizer_state_dict'])
-        
+
         self.best_metric = checkpoint['best_summary_loss']
         self.epoch = checkpoint['epoch'] + 1
 
         if 'scheduler_state_dict' in checkpoint:
             self.scheduler.load_state_dict(checkpoint['scheduler_state_dict'])
 
+    @staticmethod
     def load_model_weights(path, model):
         """
         Static method that loads weights into a model
-
         Parameters
         ----------
         path : str
             path containing the weights. Normally a .bin file
         model : torch.nn.Module
             Module to load the weights on
-
         Returns
         -------
         torch.nn.Module
-            The inputed model with loaded weights
+            The input model with loaded weights
         """
 
         checkpoint = torch.load(path)
@@ -492,7 +510,6 @@ class TorchFitter:
     def log(self, message):
         """
         Log training ouput into console and file
-
         Parameters
         ----------
         message : str
@@ -500,102 +517,53 @@ class TorchFitter:
         """
         if self.verbose:
             print(message)
-        with open(self.log_path, 'a+') as logger:
-            logger.write(f'{message}\n')
+
+        if self.save_log is True:
+            if not os.path.exists(self.base_dir):
+                os.makedirs(self.base_dir)
+            with open(self.log_path, 'a+') as logger:
+                logger.write(f'{message}\n')
 
 
-class TorchFitterBoxes(TorchFitter):
+class AutoencoderFitter(TorchFitterBase):
 
-    def validation(self, val_loader):
-        """
-        Validates a model
+    def _unpack(self, data):
+        # extract x and y from the dataloader
+        x = data['x']
 
-        Parameters
-        ----------
-        val_loader : torch.utils.data.DataLoader
-            Validation Data
-        metric : function with (y_true, y_pred, **metric_kwargs) signature
-            Metric to evaluate results on
-        metric_kwargs : dict
-            Arguments for the passed metric. Ignored if metric is None
+        # send them to device
+        x = x.to(self.device)
+        x = x.float()
 
-        Returns
-        -------
-        AverageMeter
-            object with this epochs's average loss
-        float
-            calculated metric if a metric is provided, else None
-        """
-        self.model.eval()
-        summary_loss = AverageMeter()
-        t = time.time()
-        for step, (images, targets, image_ids) in enumerate(val_loader):
-            if self.verbose:
-                if step % self.verbose == 0:
-                    print(
-                        f'\rVal Step {step}/{len(val_loader)}, '
-                        f'summary_loss: {summary_loss.avg:.5f}, '
-                        f'time: {(time.time() - t):.5f}', end=''
-                    )
-            with torch.no_grad():  # no gradient update
-                images = torch.stack(images)
-                batch_size = images.shape[0]
-                images = images.to(self.device).float()
-                boxes = [target['boxes'].to(self.device).float() for target in targets]
-                labels = [target['labels'].to(self.device).float() for target in targets]
+        # calculate batch size
+        batch_size = x.shape[0]
 
-                # just forward propagation
-                loss, _, _ = self.model(images, boxes, labels)
-                summary_loss.update(loss.detach().item(), batch_size)
+        if 'w' in data:
+            w = data['w']
+            w = w.to(self.device)
+            w = w.float()
+        else:
+            w = None
 
-        return summary_loss
+        return x, x, w, batch_size
 
-    def train_one_epoch(self, train_loader):
-        """
-        Run one epoch on the train dataset
 
-        Parameters
-        ----------
-        train_loader : torch.DataLoader
-            DataLoaders containing the training dataset
+class TransformersFitter(TorchFitterBase):
 
-        Returns
-        -------
-        AverageMeter
-            object with this epochs's average loss
-        """
-        self.model.train()  # set train mode
-        summary_loss = AverageMeter()  # object to track the average loss
-        t = time.time()
+    def _unpack(self, data):
+        x = data['x']
+        inputs, masks = x['input_ids'], x['attention_mask']
+        inputs = inputs.squeeze().to(self.device).int()
+        masks = masks.squeeze().to(self.device).int()
 
-        # run epoch
-        for step, (images, targets, image_ids) in enumerate(train_loader):
-            if self.verbose > 0:
-                if step % self.verbose == 0:
-                    print(
-                        f'\rTrain Step {step}/{len(train_loader)}, ' +
-                        f'summary_loss: {summary_loss.avg:.5f}, ' +
-                        f'time: {(time.time() - t):.5f}', end=''
-                    )
-            # extract images, boxes and labels from the dataloader
-            images = torch.stack(images)
-            images = images.to(self.device).float()
-            batch_size = images.shape[0]
-            boxes = [target['boxes'].to(self.device).float() for target in targets]
-            labels = [target['labels'].to(self.device).float() for target in targets]
+        y = data['y'].to(self.device)
+        batch_size = inputs.shape[0]
 
-            self.optimizer.zero_grad()
+        if 'w' in data:
+            w = data['w']
+            w = w.to(self.device)
+            w = w.float()
+        else:
+            w = None
 
-            loss, _, _ = self.model(images, boxes, labels)
-            # print(loss)
-
-            loss.backward()
-
-            summary_loss.update(loss.detach().item(), batch_size)
-
-            self.optimizer.step()
-
-            if self.step_scheduler:
-                self.scheduler.step()
-
-        return summary_loss
+        return [inputs, masks], y, w, batch_size
